@@ -5,89 +5,41 @@ import time
 import requests
 
 from requests import HTTPError
-from requests.compat import quote
+from requests.compat import quote, json
 from requests.auth import HTTPBasicAuth
 
 
 #-----------------------------------------------------------------------------
-__all__ = 'Job', 'Jenkins', 'Server', 'JenkinsError', 'Build', 'View'
-
+__all__ = (
+    'Job',
+    'Jenkins',
+    'Server',
+    'JenkinsError',
+    'Build',
+    'View',
+    'Node'
+)
 
 #-----------------------------------------------------------------------------
-class Job(object):
-    '''Represents a Jenkins job.'''
-
-    __slots__ = 'name', 'server'
-
-    def __init__(self, name, server):
-        self.name = name
-        self.server = server
-
-    def __str__(self):
-        return 'job:%r' % (self.name)
-
-    def __repr__(self):
-        cls = self.__class__.__name__
-        return '%s(%r)' % (cls, self.name)
-
-    def _not_exist_raise(self):
-        if not self.exists:
-            raise JenkinsError('job "%s" does not exist' % self.name)
-
-    def delete(self):
-        '''Permanently remove job.'''
-        self._not_exist_raise()
-        url = 'job/%s/doDelete' % quote(self.name)
-
-        res = self.server.post(url, throw=False)
-        if self.exists:
-            raise JenkinsError('delete of job "%s" failed' % self.name)
-        return res
-
-    def enable(self):
-        '''Enable job.'''
-        self._not_exist_raise()
-        url = 'job/%s/enable' % quote(self.name)
-        return self.server.post(url)
-
-    def disable(self):
-        '''Disable job.'''
-        self._not_exist_raise()
-        url = 'job/%s/disable' % quote(self.name)
-        return self.server.post(url)
-
-    def build(self, parameters=None, token=None):
-        '''Trigger a build.'''
-        self._not_exist_raise()
-
-        params = {}
-        if token:
-            params['token'] = token
-
-        if parameters:
-            params.update(parameters)
-            url = 'job/%s/buildWithParameters' % self.name
-        else:
-            url = 'job/%s/build' % self.name
-
-        return self.server.post(url, params=params)
-
-    def reconfigure(self, newconfig):
-        '''Update the config.xml of an existing job.'''
-        self._not_exist_raise()
-
-        url = 'job/%s/config.xml' % self.name
-        headers = {'Content-Type': 'text/xml'}
-        params = {'name': self.name}
-        return self.server.post(url, data=newconfig, params=params, headers=headers)
+class _JenkinsBase(object):
+    '''Base class for Jenkins objects.'''
 
     @property
-    def enabled(self):
-        return '<disabled>false</disabled>' in self.config
+    def baseurl(self):
+        raise NotImplementedError()
+
+    def url(self, path):
+        return '%s/%s' % (self.baseurl, path)
+
+    @property
+    def info(self):
+        url = self.url('api/json?depth=0')
+        err = '%s does not exist' % str(self)
+        return self.server.json(url, errmsg=err)
 
     @property
     def exists(self):
-        '''Check if job exists.'''
+        '''Check if object exists.'''
         try:
             self.info
             return True
@@ -98,18 +50,12 @@ class Job(object):
         except JenkinsError:
             return False
 
-    @property
-    def info(self):
-        url = 'job/%s/api/json?depth=0' % quote(self.name)
-        err = 'job "%s" does not exist' % self.name
-        return self.server.json(url, errmsg=err)
-
     @property
     def config(self):
-        url = 'job/%s/config.xml' % quote(self.name)
+        url = self.url('config.xml')
         res = self.server.get(url)
         if res.status_code != 200 or res.headers.get('content-type', '') != 'application/xml':
-            msg = 'fetching configuration for job "%s" did not return an xml document'
+            msg = 'fetching configuration for item "%s" did not return an xml document'
             raise JenkinsError(msg % self.name)
         return res.text
 
@@ -129,12 +75,88 @@ class Job(object):
         newconfig = newconfig.tostring(etree)
         self.reconfigure(newconfig)
 
+    def reconfigure(self, newconfig):
+        '''Update the config.xml of an existing item.'''
+        self._not_exist_raise()
+
+        url = self.url('config.xml')
+        headers = {'Content-Type': 'text/xml'}
+        params = {'name': self.name}
+        return self.server.post(url, data=newconfig, params=params, headers=headers)
+
+    def _not_exist_raise(self):
+        if not self.exists:
+            raise JenkinsError('%s does not exist' % str(self))
+
+
+#-----------------------------------------------------------------------------
+class Job(_JenkinsBase):
+    '''Represents a Jenkins job.'''
+
+    __slots__ = 'name', 'server'
+
+    def __init__(self, name, server):
+        self.name = name
+        self.server = server
+
+    def __str__(self):
+        return 'job:%r' % (self.name)
+
+    def __repr__(self):
+        cls = self.__class__.__name__
+        return '%s(%r)' % (cls, self.name)
+
+    @property
+    def baseurl(self):
+        return 'job/%s' % quote(self.name)
+
+    def delete(self):
+        '''Permanently remove job.'''
+        self._not_exist_raise()
+        url = self.url('doDelete')
+        res = self.server.post(url, throw=False)
+        if self.exists:
+            raise JenkinsError('delete of job "%s" failed' % self.name)
+        return res
+
+    def enable(self):
+        '''Enable job.'''
+        self._not_exist_raise()
+        url = self.url('enable')
+        return self.server.post(url)
+
+    def disable(self):
+        '''Disable job.'''
+        self._not_exist_raise()
+        url = self.url('disable')
+        return self.server.post(url)
+
+    def build(self, parameters=None, token=None):
+        '''Trigger a build.'''
+        self._not_exist_raise()
+
+        params = {}
+        if token:
+            params['token'] = token
+
+        if parameters:
+            params.update(parameters)
+            url = self.url('buildWithParameters')
+        else:
+            url = self.url('build')
+
+        return self.server.post(url, params=params)
+
+    @property
+    def enabled(self):
+        return not '<disabled>true</disabled>' in self.config
+
     @property
     def builds(self):
         return [Build(self, i['number']) for i in self.info['builds']]
 
     def __last_build_helper(self, path):
-        url = 'job/%s/%s/api/json' % (quote(self.name), path)
+        url = self.url(path + '/api/json')
         res = self.server.json(url)
         return Build(self, res['number'])
 
@@ -199,7 +221,7 @@ class Job(object):
 
 
 #-----------------------------------------------------------------------------
-class View(object):
+class View(_JenkinsBase):
     '''Represents a Jenkins view.'''
 
     __slots__ = 'name', 'server'
@@ -215,73 +237,19 @@ class View(object):
         cls = self.__class__.__name__
         return '%s(%r)' % (cls, self.name)
 
-    def _not_exist_raise(self):
-        if not self.exists:
-            raise JenkinsError('view "%s" does not exist' % self.name)
-
     @property
-    def exists(self):
-        '''Check if view exists.'''
-        try:
-            self.info
-            return True
-        except HTTPError as e:
-            if e.response.status_code == 404:
-                return False
-            raise
-        except JenkinsError:
-            return False
-
-    @property
-    def info(self):
-        url = 'view/%s/api/json?depth=0' % quote(self.name)
-        err = 'view "%s" does not exist' % self.name
-        return self.server.json(url, errmsg=err)
-
-    @property
-    def config(self):
-        url = 'view/%s/config.xml' % quote(self.name)
-        res = self.server.get(url)
-        if res.status_code != 200 or res.headers.get('content-type', '') != 'application/xml':
-            msg = 'fetching configuration for view "%s" did not return an xml document'
-            raise JenkinsError(msg % self.name)
-        return res.text
-
-    @config.setter
-    def config(self, newconfig):
-        self.reconfigure(newconfig)
-
-    @property
-    def config_etree(self):
-        # The cost of `'lxml' in sys.modules' is negligible and is
-        # preferable to having a hard dependency on lxml.
-        from lxml import etree
-        return etree.fromstring(self.config.encode('utf8'))
-
-    @config_etree.setter
-    def config_etree(self, newconfig):
-        newconfig = newconfig.tostring(etree)
-        self.reconfigure(newconfig)
+    def baseurl(self):
+        return 'view/%s' % quote(self.name)
 
     def delete(self):
         '''Permanently remove view.'''
         self._not_exist_raise()
-        url = 'view/%s/doDelete' % quote(self.name)
-
+        url = self.url('doDelete')
         res = self.server.post(url, throw=False)
         if self.exists:
             raise JenkinsError('delete of view "%s" failed' % self.name)
         return res
 
-    def reconfigure(self, newconfig):
-        '''Update the config.xml of an existing view.'''
-        self._not_exist_raise()
-
-        url = 'view/%s/config.xml' % self.name
-        headers = {'Content-Type': 'text/xml'}
-        params = {'name': self.name}
-        return self.server.post(url, data=newconfig, params=params, headers=headers)
-
     def remove_job(self, job):
         '''Remove job from view.'''
 
@@ -291,7 +259,7 @@ class View(object):
         if not job.exists:
             raise JenkinsError('job "%s" does not exist' % job.name)
 
-        url = 'view/%s/removeJobFromView' % self.name
+        url = self.url('removeJobFromView')
         params = {'name': job.name}
         res = self.server.post(url, params=params)
 
@@ -308,7 +276,7 @@ class View(object):
         if not job.exists:
             raise JenkinsError('job "%s" does not exist' % job.name)
 
-        url = 'view/%s/addJobToView' % self.name
+        url = self.url('addJobToView')
         params = {'name': job.name}
         res = self.server.post(url, params=params)
 
@@ -345,20 +313,110 @@ class View(object):
 
 
 #-----------------------------------------------------------------------------
-class Build(object):
+class NodeLaunchMethod:
+    COMMAND = 'hudson.slaves.CommandLauncher'
+    JNLP    = 'hudson.slaves.JNLPLauncher'
+    SSH     = 'hudson.plugins.sshslaves.SSHLauncher'
+    WINDOWS_SERVICE = 'hudson.os.windows.ManagedWindowsServiceLauncher'
+
+class Node(_JenkinsBase):
+    '''Represents a Jenkins node.'''
+
+    __slots__ = 'name', 'server'
+
+    def __init__(self, name, server):
+        self.name = name
+        self.server = server
+
+    def __str__(self):
+        return '<node:%s>' % (self.name)
+
+    @property
+    def baseurl(self):
+        return 'computer/%s' % quote(self.name)
+
+    @classmethod
+    def create(cls, name, remotefs, server,
+               num_executors=2,
+               node_description=None,
+               labels=None,
+               exclusive=False,
+               launcher=NodeLaunchMethod.COMMAND,
+               launcher_params={}):
+        '''
+        :param name: name of node to create, ``str``
+        :param remotefs: Remote root directory, ``str``
+        :param num_executors: number of executors for node, ``int``
+        :param node_description: Description of node, ``str``
+        :param labels: Labels to associate with node, ``str``
+        :param exclusive: Use this node for tied jobs only, ``bool``
+        :param launcher: Slave launch method, ``NodeLaunchMethod|str``
+        :param launcher_params: Additional launcher parameters, ``dict``
+        '''
+        node = cls(name, server)
+        if node.exists:
+            raise JenkinsError('node "%s" already exists' % name)
+
+        mode = 'EXCLUSIVE' if exclusive else 'NORMAL'
+        launcher_params['stapler-class'] = launcher
+
+        inner_params = {
+            'name': name,
+            'nodeDescription': node_description,
+            'numExecutors': num_executors,
+            'remoteFS': remotefs,
+            'labelString': labels,
+            'mode': mode,
+            'type': 'hudson.slaves.DumbSlave$DescriptorImpl',
+            'retentionStrategy': {
+                'stapler-class':
+                'hudson.slaves.RetentionStrategy$Always'
+            },
+            'nodeProperties': {'stapler-class-bag': 'true'},
+            'launcher': launcher_params
+        }
+
+        params = {
+            'name': name,
+            'type': 'hudson.slaves.DumbSlave$DescriptorImpl',
+            'json': json.dumps(inner_params)
+        }
+
+        res = server.post('computer/doCreateItem', params=params, throw=False)
+
+        if not res or res.status_code != 200:
+            print(res.text)
+            raise JenkinsError('create "%s" failed' % name)
+        else:
+            res.raise_for_status()
+
+    def delete(self):
+        '''Permanently remove node.'''
+        self._not_exist_raise()
+        url = self.url('doDelete')
+        res = self.server.post(url, throw=False)
+        if self.exists:
+            raise JenkinsError('delete of node "%s" failed' % self.name)
+        return res
+
+    def reconfigure(self, newconfig):
+        raise NotImplementedError
+
+
+#-----------------------------------------------------------------------------
+class Build(_JenkinsBase):
     '''Represents a Jenkins build.'''
 
-    __slots__ = 'job', 'number'
+    __slots__ = 'job', 'number', 'server'
 
     def __init__(self, job, number):
         self.job = job
         self.number = number
+        self.server = self.job.server
 
     @property
-    def info(self):
-        '''Get information about this build.'''
-        url = 'job/%s/%d/api/json' % (self.job.name, self.number)
-        return self.job.server.json(url, 'unable to retrieve info')
+    def baseurl(self):
+        return '%s/%d' % (self.job.baseurl, self.number)
 
     @property
     def building(self):
@@ -369,8 +427,8 @@ class Build(object):
         return '%s(%r, %r)' % (cls, self.job, self.number)
 
     def stop(self):
-        url = 'job/%s/%d/stop' % (self.job.name, self.number)
-        return self.job.server.post(url)
+        url = self.url('stop')
+        return self.server.post(url)
 
     def wait(self, tick=1, timeout=None):
         '''Wait for build to complete.'''
@@ -437,14 +495,22 @@ class Jenkins(object):
 
     @property
     def jobs(self):
-        return [Job(i['name'], self.server) for i in self.info['jobs']]
+        return list(self.xjobs)
+
+    @property
+    def xjobs(self):
+        return (Job(i['name'], self.server) for i in self.info['jobs'])
 
     @property
     def jobnames(self):
-        return [i['name'] for i in self.info['jobs']]
+        return list(self.xjobnames)
+
+    @property
+    def xjobnames(self):
+        return (i['name'] for i in self.info['jobs'])
 
     #-------------------------------------------------------------------------
-    # alternative job, view and build api
+    # alternative jenkins object api
     def job(self, name):
         return Job(name, self.server)
 
@@ -455,6 +521,9 @@ class Jenkins(object):
         name = name.name if isinstance(name, Job) else name
         return Build(name, number)
 
+    def node(self, name):
+        return Node(name, self.server)
+
     #-------------------------------------------------------------------------
     def job_info(self, name):
         return self.job(name).info
@@ -562,6 +631,25 @@ class Jenkins(object):
     def view_create(self, name, config):
         return View.create(name, config, self.server)
 
+    #-------------------------------------------------------------------------
+    def node_exists(self, name):
+        return Node(name, self.server).exists
+
+    def node_create(self, name, remotefs, *args, **kw):
+        return Node.create(name, remotefs, self.server, *args, **kw)
+
+    def node_info(self, name):
+        return Node(name, self.server).info
+
+    def node_delete(self, name):
+        return Node(name, self.server).delete()
+
+    def node_config(self, name):
+        return self.node(name).config
+
+    def node_config_etree(self, name):
+        return self.node(name).config_etree
+
     job_exists.__doc__ = Job.exists.__doc__
     job_delete.__doc__ = Job.delete.__doc__
     job_enable.__doc__ = Job.enable.__doc__
@@ -581,6 +669,10 @@ class Jenkins(object):
     view_delete.__doc__ = View.delete.__doc__
     view_has_job.__doc__ = View.has_job.__doc__
 
+    node_exists.__doc__ = Node.exists.__doc__
+    node_create.__doc__ = Node.create.__doc__
+    node_delete.__doc__ = Node.delete.__doc__
+
 
 #-----------------------------------------------------------------------------
 class JenkinsError(Exception):
